import cohere
import time
import uuid
from AIBridge.exceptions import CohereException, AIBridgeException, ValidationException
from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.output_validation.active_validator import ActiveValidator
import json
from AIBridge.constant.common import parse_fromat, parse_api_key
from AIBridge.output_validation.convertors import FromJson, IntoJson
from AIBridge.constant.common import get_function_from_json


class CohereApi(AIInterface):
    """
    Base class for OpenAI Services
    """

    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        output_format: list[str] = [],
        format_strcture: list[str] = [],
        model="command-nightly",
        variation_count: int = 1,
        max_tokens: int = 3500,  # max 4096
        temperature: float = 0.5,
        message_queue=False,
        api_key=None,
        output_format_parse=True,
        context=[],
    ):
        try:
            if prompts and prompt_ids:
                raise CohereException(
                    "please provide either prompts or prompts ids at time"
                )
            if not prompts and not prompt_ids:
                raise CohereException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if output_format:
                if len(output_format) != len(prompts_list):
                    raise ValidationException(
                        "length of output_format must be equal to length of the prompts"
                    )
            if format_strcture:
                if len(format_strcture) != len(prompts_list):
                    raise ValidationException(
                        "length of format_strcture must be equal to length of the prompts"
                    )
            updated_prompts = []
            for _prompt in prompts_list:
                format = None
                format_str = None
                if output_format:
                    format = output_format[prompts_list.index(_prompt)]
                if format_strcture:
                    format_str = format_strcture[prompts_list.index(_prompt)]
                if output_format_parse:
                    u_prompt = parse_fromat(
                        _prompt, format=format, format_structure=format_str
                    )
                    updated_prompts.append(u_prompt)
            if not updated_prompts:
                updated_prompts = prompts_list
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(updated_prompts),
                    "model": model,
                    "variation_count": variation_count,
                    "max_tokens": max_tokens,
                    "temperature": temperature,
                    "ai_service": "cohere_api",
                    "output_format": json.dumps(output_format),
                    "format_structure": json.dumps(format_strcture),
                    "api_key": api_key,
                    "context": context,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                updated_prompts,
                model,
                variation_count,
                max_tokens,
                temperature,
                output_format,
                format_strcture,
                api_key=api_key,
                context=context,
            )
        except Exception as e:
            raise CohereException(e)

    @classmethod
    def get_prompt_context(self, context):
        context_list = []
        prev_role = ""
        if context:
            for _context in context:
                if _context["role"] not in ["user", "system", "assistant"]:
                    raise CohereException(
                        "Invalid role provided. Please provide either user or system, assistant"
                    )
                key = "USER"
                if _context["role"] in ["assistant", "system"]:
                    key = "CHATBOT"
                if prev_role == key:
                    if key == "USER":
                        context_list.append({"role": "CHATBOT", "text": "understood"})
                    elif key == "CHATBOT":
                        context_list.append({"role": "USER", "text": "generated"})
                else:
                    context_list.append({"role": key, "text": _context["context"]})
                prev_role = key
            if context_list:
                data_test = context_list[-1]
                if data_test["role"] == "USER":
                    context_list.append({"role": "CHATBOT", "text": "understood"})
        return context_list

    @classmethod
    def execute_text_prompt(
        self, api_key, model, messages, n, max_tokens=3500, temperature=0.5, context=[]
    ):
        co = cohere.Client(api_key)
        return co.chat(
            model=model,
            chat_history=context,
            message=messages,
            max_tokens=max_tokens,
            temperature=temperature,
        )

    @classmethod
    def execute_prompt_function_calling(
        self,
        api_key,
        model,
        messages,
        n,
        functions_call,
        max_tokens=3500,
        temperature=0.5,
        context=[],
    ):
        co = cohere.Client(api_key)
        return co.chat(
            model=model,
            chat_history=context,
            message=messages,
            tools=functions_call,
            max_tokens=max_tokens,
            temperature=temperature,
        )

    @classmethod
    def get_response(
        self,
        prompts,
        model="command-nightly",
        variation_count=1,
        max_tokens=3500,  # max 4096
        temperature=0.5,
        output_format=[],
        format_structure=[],
        api_key=None,
        context=[],
    ):
        try:
            if output_format:
                if isinstance(output_format, str):
                    output_format = json.loads(output_format)
            if format_structure:
                if isinstance(format_structure, str):
                    format_structure = json.loads(format_structure)
            if not prompts:
                raise CohereException("No prompts provided")
            print(prompts)
            api_key = api_key if api_key else parse_api_key("cohere_api")
            model_output = []
            message_data = []
            context = self.get_prompt_context(context)
            message_data=context
            token_used = 0
            input_tokens = 0
            output_tokens = 0
            _formatter = "string"
            for index, prompt in enumerate(prompts):
                if output_format:
                    _formatter = output_format[index]
                if _formatter not in ["json", "csv", "xml"]:
                    response = self.execute_text_prompt(
                        api_key,
                        model=model,
                        messages=prompt,
                        n=variation_count,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        context=message_data,
                    )
                    message_data.append(
                        {
                            "role": "CHATBOT",
                            "content": response.text,
                        }
                    )
                    content = response.text
                else:
                    if _formatter == "csv":
                        csv=format_structure[index]
                        if ";" in csv:
                            csv = csv.replace(";", ",")
                        schema = IntoJson.csv_to_json(csv)
                    elif _formatter == "xml":
                        schema = IntoJson.xml_to_json(format_structure[index])
                    elif _formatter == "json":
                        schema = json.loads(format_structure[index])
                    print(schema,"ssssssssssssssssss")
                    functions = [get_function_from_json(schema, call_from="cohere")]
                    response = self.execute_prompt_function_calling(
                        api_key=api_key,
                        model=model,
                        messages=prompt,
                        n=variation_count,
                        functions_call=functions,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        context=message_data,
                    )
                    content = response.tool_calls[0].parameters
                tokens = (
                    response.meta["billed_units"]["input_tokens"]
                    + response.meta["billed_units"]["output_tokens"]
                )
                input_tokens += response.meta["billed_units"]["input_tokens"]
                output_tokens += response.meta["billed_units"]["output_tokens"]
                token_used = token_used + tokens
                if output_format:
                    if isinstance(content, dict):
                        content = json.dumps(content)
                    try:
                        if _formatter == "csv":
                            print(content)
                            if isinstance(content, str):
                                content = json.loads(content)
                            content = FromJson.json_to_csv(content)
                        elif _formatter == "xml":
                            if isinstance(content, str):
                                content = json.loads(content)
                            content = FromJson.json_to_xml(content)
                    except AIBridgeException as e:
                        raise ValidationException(
                            f"Ai response is not in valid {_formatter}"
                        )
                    if _formatter == "json":
                        _validate_obj = ActiveValidator.get_active_validator(_formatter)
                        try:
                            content = _validate_obj.validate(
                                content,
                                schema=(
                                    format_structure[index]
                                    if format_structure
                                    else None
                                ),
                            )
                        except AIBridgeException as e:
                            content_error = {
                                "error": f"{e}",
                                "ai_response": content,
                            }
                            content = json.dumps(content_error)
                if index >= len(model_output):
                    model_output.append({"data": [content]})
                else:
                    model_output[index]["data"].append(content)
            message_value = {
                "items": {
                    "response": model_output,
                    "input_tokens": input_tokens,
                    "output_tokens": output_tokens,
                    "token_used": token_used,
                    "created_at": time.time(),
                    "ai_service": "cohere",
                }
            }
            return message_value
        except Exception as e:
            raise CohereException(e)



"""
Exception  handling
"""

import json


class AIBridgeException(Exception):
    def __init__(self, message, code=400):
        if isinstance(message, str):
            message = message
        elif isinstance(message, dict):
            code = message.get("code", 0)
            if code == 0:
                code = message.get("status_code", 0)
            if code == 0:
                code = message.get("error_code", 400)
            if isinstance(message, dict):
                if "error" in message:
                    if "message" in message["error"]:
                        message = message["error"]["message"]
                    else:
                        message = str(message)
                elif "details" in message:
                    message = message["details"]
                    if "detail" in message:
                        message = message["detail"]
        else:
            if hasattr(message, "status_code"):
                code = message.status_code
            elif hasattr(message, "code"):
                code = message.code
            elif hasattr(message, "error_code"):
                code = message.error_code
            else:
                code = 400
            if hasattr(message, "response"):
                message = message.response
                if isinstance(message, object):
                    if hasattr(message, "message"):
                        message = message.message
                    elif hasattr(message, "json"):
                        message = message.json()
                    elif hasattr(message, "details"):
                        message = message.details
                        if isinstance(message, object):
                            if hasattr(message, "details"):
                                message = message.details
                    elif hasattr(message, "text"):
                        message = message.text
            elif hasattr(message, "body"):
                message = message.body
            elif hasattr(message, "error"):
                message = message.error
            elif hasattr(message, "details"):
                message = message.details
            elif hasattr(message, "message"):
                message = message.message
            print(code, message, "xxxxxxxxxxxxxx")
            if not isinstance(message, object):
                try:
                    message = json.loads(message)
                except json.JSONDecodeError as e:
                    message = str(message)
            if isinstance(message, dict):
                if "error" in message:
                    if "message" in message["error"]:
                        message = message["error"]["message"]
                    else:
                        message = str(message)
                elif "message" in message:
                    message = message["message"]
                
                elif "detail" in message:
                    print(message,"mmmmmmmmmmmmmmmmmmmmm]mmmmmmm",type(message))
                    message = message["detail"]
            else:
                message = str(message)
        self.message = message
        self.code = code

    def __str__(self):
        return f"status_code: {self.code}, message: {self.message}"


class OpenAIException(AIBridgeException):
    def __init__(self, message):
        print(message)
        super().__init__(message)


class ConfigException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class ValidationException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class PromptSaveException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class VariablesException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class PromptCompletionException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class AssignQueueException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class ProcessMQException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class AIResponseException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class MessageQueueException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class DatabaseException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class ImageException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class PalmTextException(AIBridgeException):
    def __init__(self, message):
        self.message = message
        self.code = 1000


class StableDiffusionException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class CohereException(AIBridgeException):
    def __init__(self, message):
        print(message, "xxxxxxxxxxxxxxxxxxxxx")
        super().__init__(message)


class Ai21Exception(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class GeminiException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class AnthropicsException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)


class AiBridgeValidationException(AIBridgeException):
    def __init__(self, message):
        super().__init__(message)



from AIBridge.constant.constant import (
    COHERE_FUNCTION_CALL_FORMAT,
    FUNCTION_CALL_FORMAT,
    PRIORITY,
)
from AIBridge.setconfig import SetConfig
from AIBridge.exceptions import ConfigException
from AIBridge.database.sql_service import SQL
from AIBridge.database.no_sql_service import Mongodb
from AIBridge.exceptions import AIBridgeException
from urllib.parse import urlparse
from AIBridge.output_validation.convertors import FromJson, IntoJson
import json
from PIL import Image
import requests
import base64
import io

config = SetConfig.read_yaml()


def parse_fromat(prompt, format=None, format_structure=None):
    if format:
        prompt = prompt + f"format:json valid"
    if format_structure:
        if format:
            if format == "csv":
                if ";" in format_structure:
                    format_structure = format_structure.replace(";", ",")
                format_structure = json.dumps(IntoJson.csv_to_json(format_structure))
            elif format == "xml":
                format_structure = json.dumps(IntoJson.xml_to_json(format_structure))

        prompt = prompt + f"format_structure:{format_structure}"
    prompt = (
        prompt
        + " Respond only in the exact specified  json format provided in the prompt,No extra information,No extra space please provide valid data"
    )
    return prompt


def parse_api_key(ai_service):
    if ai_service not in config:
        raise ConfigException("ai_service not found in config file")
    for ext in PRIORITY:
        if ext in config[ai_service]:
            if config[ai_service][ext]:
                return config[ai_service][ext][0]
    raise ConfigException(f" {ai_service} api_key not found in config file")


def get_no_sql_obj():
    databse_uri = config["database_uri"]
    if "mongodb" in databse_uri:
        return Mongodb()


def get_database_obj():
    if "database" not in config:
        return SQL()
    elif config["database"] == "nosql":
        return get_no_sql_obj()
    elif config["database"] == "sql":
        return SQL()


def get_function_from_json(output_schema: dict, call_from="open_ai"):
    type_ = "type"
    object = "object"
    array = "array"
    string = "string"
    if call_from == "gemini_ai":
        type_ = "type_"
        object = "OBJECT"
        array = "ARRAY"
        string = "STRING"
    elif call_from == "cohere":
        type_ = "type"
        object = "object"
        array = "array"
        string = "str"

    def create_function_call(output_schema: dict):
        if call_from == "cohere":
            # properties = "parameter_definitions"
            properties = "properties"
        else:
            properties = "properties"
        key_dict = {type_: object, properties: {}}
        for key, value in output_schema.items():
            if isinstance(value, dict):
                key_dict[properties][key] = create_function_call(value)
            elif isinstance(value, list):
                key_d = {
                    type_: array,
                    "description": "Generate the information",
                }
                if call_from == "cohere":
                    key_d["required"] = True
                if value:
                    if isinstance(value[0], str):
                        key_d["items"] = {type_: string, "description": value[0]}
                    elif isinstance(value[0], dict):
                        key_d["items"] = create_function_call(value[0])
                else:
                    key_d["items"] = {
                        type_: string,
                        "description": "provide the given information",
                    }
                    if call_from == "cohere":
                        key_d["required"] = True
                key_dict[properties][key] = key_d
            else:
                val_x = {type_: string, "description": value}
                if call_from == "cohere":
                    val_x["required"] = True
                key_dict[properties][key] = val_x
        return key_dict

    required = []
    for key in output_schema.keys():
        required.append(key)
    key_dict = create_function_call(output_schema)
    if call_from == "open_ai" or call_from == "gemini_ai":
        data = FUNCTION_CALL_FORMAT
        data["parameters"] = key_dict
        if call_from == "open_ai":
            data["parameters"]["required"] = required
    elif call_from == "cohere":
        data = COHERE_FUNCTION_CALL_FORMAT
        data["parameter_definitions"] = key_dict["properties"]
    return data



def check_url(url_list: list):
    byte_list=[]
    for url in url_list:
        if isinstance(url, str) and url.startswith(('http://', 'https://')):
            response = requests.get(url)
            image_bytes = io.BytesIO(response.content)
    
    # Check if the input is Base64
        elif isinstance(url, str) and url.startswith('data:image'):
            header, base64_data = url.split(',', 1)
            image_bytes = io.BytesIO(base64.b64decode(base64_data))
        
        # If the input is bytes, use it directly
        elif isinstance(url, (bytes, bytearray)):
            image_bytes = io.BytesIO(url)
        
        else:
            raise ValueError("Input must be a URL, base64 string, or bytes.")
        try:
            image = Image.open(image_bytes)
            format = image.format.lower()
            if format not in ["png", "jpeg", "webp"]:
                raise AIBridgeException("Invalid image format")
        except Exception as e:
            raise AIBridgeException(f"{e} : image should be one of png,jpeg,webp")
        image_bytes.seek(0)
        byte_list.append(image_bytes.read())
        image_bytes.close()
    return byte_list
    


class AnthropicsFunctionCall:
    @classmethod
    def construct_format_parameters_prompt(self, parameters):
        def create_function_call(parameter):
            key_dict = f"<parameter>\n"
            for key, value in parameter.items():
                if isinstance(value, dict):
                    j_type = "object"
                    key_dict += (
                        f"<name>{key}</name>\n<type>{j_type}</type>\n"
                        + create_function_call(value)
                    )
                elif isinstance(value, list):
                    type_ = "array"
                    key_dict += f"<name>{key}</name>\n<type>{type_}</type>"

                    if value:
                        if isinstance(value[0], str):
                            key_dict += f"\n<description>{value[0]}</description>\n"
                        elif isinstance(value[0], dict):
                            key_dict += create_function_call(value[0])
                    else:
                        description = "provide the given information"
                        key_dict += f"\n<description>{description}</description>\n"
                else:
                    type_ = "str"
                    key_dict = f"<name>{key}</name>\n<type>{type_}</type>\n<description>{value}</description>\n"
            key_dict += "</parameter>"
            return key_dict

        key_dict = create_function_call(parameters)
        return key_dict

    @classmethod
    def construct_tool_use_system_prompt(self, tools):
        tool_use_system_prompt = (
            "In this environment you have access to a set of tools you can use to answer the user's question.\n"
            "\n"
            "You may call them like this:\n"
            "<function_calls>\n"
            "<invoke>\n"
            "<tool_name>$TOOL_NAME</tool_name>\n"
            "<parameters>\n"
            "<$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>\n"
            "...\n"
            "</parameters>\n"
            "</invoke>\n"
            "</function_calls>\n"
            "\n"
            "Here are the tools available:\n"
            "<tools>\n" + "\n".join([tool for tool in tools]) + "\n</tools>"
        )
        return tool_use_system_prompt

    @classmethod
    def construct_format_tool_for_claude_prompt(
        self,
        name="user_output",
        description="provide correct user_output",
        parameters=None,
    ):
        constructed_prompt = (
            "<tool_description>\n"
            f"<tool_name>{name}</tool_name>\n"
            "<description>\n"
            f"{description}\n"
            "</description>\n"
            "<parameters>\n"
            f"{self.construct_format_parameters_prompt(parameters)}\n"
            "</parameters>\n"
            "</tool_description>"
        )
        system_prompt = self.construct_tool_use_system_prompt([constructed_prompt])
        return system_prompt



import json
import csv
import xmltodict
import io
import re


class IntoJson:
    @classmethod
    def csv_to_json(self, csv_data):
        csv_data = csv_data.replace("\n", "")
        csv_list = csv_data.split(",")
        json_dict = {"data": []}
        new_dict = {}
        for key in csv_list:
            new_dict[key] = ""
        json_dict["data"].append(new_dict)
        dict_=json_dict["data"][0]
        return dict_

    @classmethod
    def xml_to_json(self, xml_data):
        if """<?xml version="1.0" encoding="UTF-8"?>""" in xml_data:
            xml_data = xml_data.replace(
                """<?xml version="1.0" encoding="UTF-8"?>""", ""
            )

        cleaned_xml_data = re.sub(r"^.*?(<[^>]*>)", r"\1", xml_data, flags=re.DOTALL)
        json_data = xmltodict.parse(cleaned_xml_data)
        new_json = {}
        for key, value in json_data.items():
            new_json[key] = [value]
            break
        return new_json


class FromJson:
    @classmethod
    def json_to_csv(self, json_dict):
        # Convert JSON to CSV
        if "data" in json_dict:
            json_data = json_dict["data"][0] 
        else:
            json_data=json_dict
        print(json_dict,"xxxxxxxxxxxxxxxx")
        csv_output = io.StringIO()
        fieldnames= json_data.keys if isinstance(json_data,dict) else json_data[0].keys() 
        csv_writer = csv.DictWriter(csv_output, fieldnames=fieldnames)
        csv_writer.writeheader()
        json_data=json_data if isinstance(json_data,list) else [json_data]
        csv_writer.writerows(json_data)
        # Get the CSV data as a string
        csv_data = csv_output.getvalue()
        print(csv_data,"xxxxxxxxxxxxxx")
        return csv_data

    @classmethod
    def json_to_xml(self, json_data):
        root_element_name = "root"
        if isinstance(json_data,list):
            wrapped_json = {root_element_name: {"item": json_data}}
        else:
            wrapped_json = {root_element_name: json_data}
        xml_data = xmltodict.unparse(wrapped_json, pretty=True)
        return xml_data



from AIBridge.output_validation.convertors import FromJson, IntoJson
import time
import uuid
from AIBridge.exceptions import (
    AnthropicsException,
    AIBridgeException,
    ValidationException,
)
from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.output_validation.active_validator import ActiveValidator
import json
from AIBridge.constant.common import (
    get_function_from_json,
    parse_fromat,
    parse_api_key,
    AnthropicsFunctionCall,
)
from anthropic import Anthropic


class AnthropicService(AIInterface):
    """
    Base class for Anthropics Services
    """

    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        output_format: list[str] = [],
        format_strcture: list[str] = [],
        model="claude-3-opus-20240229",
        variation_count: int = 1,
        max_tokens: int = 4000,
        temperature: float = 0.5,
        message_queue=False,
        api_key=None,
        output_format_parse=True,
        stream=False,
        context=[],
    ):
        try:
            if prompts and prompt_ids:
                raise AnthropicsException(
                    "please provide either prompts or prompts ids at atime"
                )
            if not prompts and not prompt_ids:
                raise AnthropicsException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if output_format:
                if len(output_format) != len(prompts_list):
                    raise ValidationException(
                        "length of output_format must be equal to length of the prompts"
                    )
            if format_strcture:
                if len(format_strcture) != len(prompts_list):
                    raise ValidationException(
                        "length of format_strcture must be equal to length of the prompts"
                    )
            updated_prompts = []
            for _prompt in prompts_list:
                format = None
                format_str = None
                if output_format:
                    format = output_format[prompts_list.index(_prompt)]
                if format_strcture:
                    format_str = format_strcture[prompts_list.index(_prompt)]
                if output_format_parse:
                    u_prompt = parse_fromat(
                        _prompt, format=format, format_structure=format_str
                    )
                    updated_prompts.append(u_prompt)
            if not updated_prompts:
                updated_prompts = prompts_list
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(updated_prompts),
                    "model": model,
                    "variation_count": variation_count,
                    "max_tokens": max_tokens,
                    "temperature": temperature,
                    "ai_service": "open_ai",
                    "output_format": json.dumps(output_format),
                    "format_structure": json.dumps(format_strcture),
                    "api_key": api_key,
                    "context": context,
                    "stream": True,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                updated_prompts,
                model,
                variation_count,
                max_tokens,
                temperature,
                output_format,
                format_strcture,
                api_key=api_key,
                context=context,
                stream=stream,
            )
        except Exception as e:
            raise AnthropicsException(e)

    @classmethod
    def execute_text_prompt(
        self, api_key, model, messages, n, max_tokens=3500, temperature=0.5
    ):
        client = Anthropic(api_key=api_key)
        return client.messages.create(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature,
        )

    @classmethod
    def execute_prompt_function_calling(
        self,
        api_key,
        model,
        messages,
        n,
        functions_call,
        max_tokens=3500,
        temperature=0.5,
    ):
        client = Anthropic(api_key=api_key)
        return client.messages.create(
            model=model,
            messages=messages,
            system=functions_call,
            max_tokens=max_tokens,
            temperature=temperature,
        )

    @classmethod
    def get_prompt_context(self, context):
        context_list = []
        prev_role = ""
        if context:
            for _context in context:
                if _context["role"] not in ["user", "system", "assistant"]:
                    raise AnthropicsException(
                        "Invalid role provided. Please provide either user or system, assistant"
                    )
                key = _context["role"]
                if key in ["system", "assistant"]:
                    key = "assistant"
                if prev_role == key:
                    if key == "user":
                        context_list.append(
                            {"role": "assistant", "content": "understood"}
                        )
                    elif key == "assistant" or key == "system":
                        context_list.append({"role": "user", "content": "thanks"})
                else:

                    context_list.append({"role": key, "content": _context["context"]})
                prev_role = key
            if context_list:
                data_test = context_list[-1]
                if data_test["role"] == "user":
                    context_list.append({"role": "assistant", "content": "understood"})
                data = context_list[0]
                if data["role"] == "assistant":
                    context_list = [
                        {"role": "user", "content": "you are friendly assistance"}
                    ] + context_list
        return context_list

    @classmethod
    def get_response(
        self,
        prompts,
        model="claude-3-opus-20240229",
        variation_count=1,
        max_tokens=3500,
        temperature=0.5,
        output_format=[],
        format_structure=[],
        api_key=None,
        context=[],
        stream=False,
    ):
        try:
            if output_format:
                if isinstance(output_format, str):
                    output_format = json.loads(output_format)
            if format_structure:
                if isinstance(format_structure, str):
                    format_structure = json.loads(format_structure)
            if not prompts:
                raise AnthropicsException("No prompts provided")
            api_key = api_key if api_key else parse_api_key("anthropic")
            message_data = self.get_prompt_context(context)
            model_output = []
            token_used = 0
            input_tokens = 0
            output_tokens = 0
            _formatter = "string"
            func_call = False
            for index, prompt in enumerate(prompts):
                if output_format:
                    _formatter = output_format[index]
                message_data.append({"role": "user", "content": prompt})
                if _formatter not in ["json", "csv", "xml"]:
                    response = self.execute_text_prompt(
                        api_key,
                        model=model,
                        messages=message_data,
                        n=variation_count,
                        max_tokens=max_tokens,
                        temperature=temperature,
                    )
                    message_data.append(
                        {
                            "role": response.role,
                            "content": response.content[0].text,
                        }
                    )
                else:
                    func_call = True
                    if _formatter == "csv":
                        schema = IntoJson.csv_to_json(format_structure[index])
                    elif _formatter == "xml":
                        schema = IntoJson.xml_to_json(format_structure[index])
                    elif _formatter == "json":
                        schema = json.loads(format_structure[index])
                    functions = (
                        AnthropicsFunctionCall.construct_format_tool_for_claude_prompt(
                            parameters=schema
                        )
                    )
                    print(functions)
                    response = self.execute_prompt_function_calling(
                        api_key=api_key,
                        model=model,
                        messages=message_data,
                        n=variation_count,
                        functions_call=functions,
                        max_tokens=max_tokens,
                        temperature=temperature,
                    )
                    message_data.append(
                        {
                            "role": response.role,
                            "content": response.content[0].text,
                        }
                    )
                tokens = response.usage.input_tokens + response.usage.output_tokens
                token_used = token_used + tokens
                input_tokens = input_tokens + response.usage.input_tokens
                output_tokens = output_tokens + response.usage.output_tokens
                for index, res in enumerate(response.content):
                    content = res.text
                    if func_call:
                        json_content = IntoJson.xml_to_json(content)
                        content = json_content["function_calls"][0]["invoke"][
                            "parameters"
                        ]
                        if isinstance(content, dict):
                            for key, value in content.items():
                                if isinstance(value, str):
                                    try:
                                        new_val = json.loads(value)
                                        content[key] = new_val
                                    except json.JSONDecodeError:
                                        ...
                            content = json.dumps(content)

                    if output_format:
                        try:
                            if _formatter == "csv":
                                content = FromJson.json_to_csv(json.loads(content))
                            elif _formatter == "xml":
                                content = FromJson.json_to_xml(json.loads(content))
                        except AIBridgeException as e:
                            raise ValidationException(
                                f"Ai response is not in valid {_formatter}"
                            )
                        if _formatter == "json":
                            _validate_obj = ActiveValidator.get_active_validator(
                                _formatter
                            )
                            try:
                                content = _validate_obj.validate(
                                    content,
                                    schema=(
                                        format_structure[index]
                                        if format_structure
                                        else None
                                    ),
                                )
                            except AIBridgeException as e:
                                content_error = {
                                    "error": f"{e}",
                                    "ai_response": content,
                                }
                                content = json.dumps(content_error)
                    if index >= len(model_output):
                        model_output.append({"data": [content]})
                    else:
                        model_output[index]["data"].append(content)
            message_value = {
                "items": {
                    "response": model_output,
                    "input_tokens": input_tokens,
                    "output_tokens": output_tokens,
                    "token_used": token_used,
                    "created_at": time.time(),
                    "ai_service": "anthropic",
                }
            }
            return message_value
        except Exception as e:
            raise AnthropicsException(e)



from abc import ABC, abstractmethod
import json


class Caller(ABC):
    @abstractmethod
    def get_response(self, service_obj, message_data):
        pass


class OpenAiRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            context=message_data["context"],
        )
        return data


class GeminiAiRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            stop_subsequence=message_data["stop_subsequence"],
            stream=message_data["stream"],
            context=message_data["context"],
        )
        return data


class AnthropicRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            stop_subsequence=message_data["stop_subsequence"],
            stream=message_data["stream"],
            context=message_data["context"],
        )
        return data


class OpenAiImageRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            image_data=json.loads(message_data["image_data"]),
            mask_image=json.loads(message_data["mask_image"]),
            variation_count=message_data["variation_count"],
            size=message_data["size"],
            process_type=message_data["process_type"],
            api_key=message_data["api_key"],
            model=message_data["api_key"],
            quality=message_data["quality"],
        )
        return data


class PalmTextRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            context=message_data["context"],
        )
        return data


class PalmChatRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            messages=message_data["messages"],
            context=message_data["context"],
            examples=message_data["examples"],
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            temperature=message_data["temperature"],
            api_key=message_data["api_key"],
        )
        return data


class StableDuffusionRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            image_data=json.loads(message_data["image_data"]),
            mask_image=json.loads(message_data["mask_image"]),
            action=message_data["action"],
            message_queue=message_data["message_queue"],
            api_key=message_data["api_key"],
        )
        return data


class CohereRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            context=message_data["context"],
        )
        return data


class JarasicTextRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
            context=message_data["context"],
        )
        return data



from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.exceptions import (
    StableDiffusionException,
    AIBridgeException,
    ValidationException,
)
import json
import uuid
from AIBridge.constant.constant import STABLE_DIFFUSION_TYPES
from AIBridge.constant.common import parse_api_key, check_url
import requests
import time


class StableDiffusion(AIInterface):
    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        image_data: list[str] = [],
        mask_image: list[str] = [],
        action="text2img",
        message_queue=False,
        model="sd3-large",
        api_key=None,
    ):
        try:
            if prompts and prompt_ids:
                raise ValidationException(
                    "please provide either prompts or prompts ids at atime"
                )
            if not prompts and not prompt_ids:
                raise ValidationException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if action not in STABLE_DIFFUSION_TYPES:
                raise ValidationException(
                    f"action should be one of the{STABLE_DIFFUSION_TYPES}"
                )
            if action == "image2image" or action == "inpaint":
                if not image_data:
                    raise ValidationException("Please enter image link in image data")
            if action == "inpaint" and not mask_image:
                raise ValidationException("Please enter mask image link in mask image")
            if mask_image:
                if len(mask_image) != len(image_data):
                    raise ValidationException(
                        "mask_image length should be equal to image_data length",
                    )
                mask_image=check_url(mask_image)
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if image_data:
                if prompts_list:
                    if len(image_data) != len(prompts_list):
                        raise ValidationException(
                            "image_data length should be equal to prompts length",
                        )
                image_data=check_url(image_data)
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(prompts_list),
                    "ai_service": "stable_diffusion",
                    "image_data": json.dumps(image_data),
                    "mask_image": json.dumps(mask_image),
                    "action": action,
                    "message_queue": message_queue,
                    "model":model,
                    "api_key": api_key,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                prompts=prompts_list,
                image_data=image_data,
                mask_image=mask_image,
                action=action,
                message_queue=message_queue,
                model=model,
                api_key=api_key,
            )
        except Exception as e:
            raise StableDiffusionException(e)

    @classmethod
    def get_response(
        self,
        prompts,
        image_data=[],
        mask_image=[],
        action="text2img",
        message_queue=False,
        model="sd3-large",
        api_key=None,
    ):
        
        if not api_key:
            api_key = api_key if api_key else parse_api_key("stable_diffusion")
        headers={
            "authorization": f"Bearer {api_key}",
            "accept": "image/*",
        }
        url = "https://api.stability.ai/v2beta/stable-image/generate/sd3"
        if action == "image2image":
            url = "https://api.stability.ai/v2beta/stable-image/generate/sd3"
        if action == "inpaint":
            url = "https://api.stability.ai/v2beta/stable-image/edit/inpaint"
        data = []
        for index, prompt in enumerate(prompts):
            payload={
                "prompt": prompt,
                "output_format": "jpeg",
            }
            if action == "img2img":
                payload["mode"]="image-to-image"
                payload["model"]=model
                payload["strength"]=0.5
                if image_data:
                    files={
                        "image": image_data[index],
                    }
            elif action == "inpaint":
                files={
                    "image": image_data[index],
                    "mask": mask_image[index],
                }
            else:
                files={"none": ''}
                payload["model"]=model
            response = requests.post(url, headers=headers,files=files, data=payload)
            if not response.content:
                res = response.json()
                if "errors" in res:
                    raise StableDiffusionException(res["errors"][0])
            res=response.content
            content_type=response.headers.get("content-type")
            data.append({"image": res,"content_type":content_type})
        message_value = {
            "items": {
                "response": data,
                "token_used": 0,
                "created_at": time.time(),
                "ai_service": "stable_diffusion",
            }
        }
        return message_value